(* $Id: gontbind.ml,v 1.8 2001/11/16 00:00:46 malekith Exp $ *)

open Gontutil

(* The purpouse of this file to make other part of compiler 
 * symbol-table-free. All occurences of Id(string) are changed
 * to Ref(symbol, int), where symbol is object type used for
 * symbols and int tells how deep in closure stack should we go. 
 *)

class symbol = fun (scope, fnc, t, n) ->
  object
    val sym_type : ty = t
    val sym_name : id = n
    val sym_scope_level : int = scope
    val sym_fnc_level : int = fnc
    method get_name = sym_name
    method get_type = sym_type
    method get_fnc_level = sym_fnc_level
    method get_scope_level = sym_scope_level
    val mutable sym_closured = false
    method in_closure = sym_closured
    method put_into_closure = sym_closured <- true
  end

type ctx = {ctx_syms: symbol list;
            ctx_fnc_level: int;
            ctx_scope_level: int}

type symbol_initializer = int * int * ty * id

let make_symbol c t n =
  new symbol (c.ctx_scope_level, c.ctx_fnc_level, t, n)
;;


let global_ctx = {ctx_syms = []; ctx_fnc_level = 0; ctx_scope_level = 0}

(* We use Hashtbl instead of list at the global level, since set of global
 * variables is expected to be bigger, and we don't need to copy it. *)
let globals : (id, symbol) Hashtbl.t = Hashtbl.create 16

let find_sym syms n =
  try
    List.find (fun sym -> sym#get_name = n) syms
  with Not_found -> 
    (* Of course this might also raise Not_found. *)
    Hashtbl.find globals n
;;

let prepare_ctx loc (c, parms) =
  let rec helper (c, parms) =
    match parms with
        (t, n)::xs -> 
          (let (fnc_lev, scope_lev) =
            try
              let sym = find_sym c.ctx_syms n in
                (sym#get_fnc_level, sym#get_scope_level)
            with Not_found -> (-1, -1) in
          if fnc_lev == c.ctx_fnc_level then
            err (loc, xf "parameter `%s' defined twice" n);
          if scope_lev >= 0 then
            warn (loc, xf "`%s' shodows other symbol" n));
          let sym = make_symbol c t n in
          helper ({c with ctx_syms = sym :: c.ctx_syms }, xs)
      | [] -> c in
  helper ({c with ctx_fnc_level   = c.ctx_fnc_level + 1;
                  ctx_scope_level = c.ctx_scope_level + 1}, 
          parms)
;;

(* This does inplace modification of s. Updated context is returned. *)
let rec bind_stmt (c, s) =
  match s.raw_stmt with
      Skip -> c
    | Exp(e) -> bind_exp (c, e)
    | VarDef v ->
          (match v.var_symbol with
              None -> ()
            | Some _ -> assert false);
          (let scope_lev =
            try
              let sym = find_sym c.ctx_syms v.var_name in
                sym#get_scope_level
            with Not_found -> -1 in
          if scope_lev == c.ctx_scope_level then
            err (s.stmt_loc, xf "variable `%s' defined twice" v.var_name);
          if scope_lev >= 0 then
            warn (s.stmt_loc, xf "`%s' shodows other symbol" v.var_name));
          let sym = make_symbol c v.var_type v.var_name in
          v.var_symbol <- Some sym;
          {c with ctx_syms = sym :: c.ctx_syms}
    | Return(e) -> bind_exp (c, e)
    | ReturnVoid -> c
    | Block(ss) ->
        let rec bind_block (c, ss) =
          match ss with
              x::xs -> 
                let c' = bind_stmt (c, x) in
                bind_block (c', xs)
            | [] -> c in
        let new_scope = {c with ctx_scope_level = c.ctx_scope_level + 1} in
        let c' = bind_block(new_scope, ss) in
        ignore(c'); 
        c
and bind_exp (c, e) =
  match e.raw_exp with
      SimpleBinOp(op, e1, e2) -> 
        let c' = bind_exp (c, e1) in
        bind_exp (c', e2)
    | SimpleUnOp(op, e1) ->
        bind_exp (c, e1)
    | Ref(r) ->
        (match r.ref_symbol with
            None ->
              let sym = 
                try
                  find_sym c.ctx_syms r.ref_name
                with Not_found ->
                  err (e.exp_loc, xf "unbound symbol `%s'" r.ref_name) in
              r.ref_symbol <- Some sym;
              let lev = c.ctx_fnc_level - sym#get_fnc_level in
                (* prerr_endline (xf "binding symbol %s (%d -> %d)" r.ref_name
                c.ctx_fnc_level sym#get_fnc_level); *)
                assert (lev >= 0);
                r.ref_level <- if sym#get_fnc_level == 0 then 0 else lev
          (* We have already been here! *)
          | Some _ -> assert false); 
          (* Context is not updated in any way. *)
          c
    | Call(e1, es) -> 
        let c' = bind_exp (c, e1) in
        let rec bind_parms (c, pp) =
            match pp with
                p::ps -> 
                  let c' = bind_exp (c, p) in
                  bind_parms (c', ps)
              | [] -> c in
        bind_parms (c', es)
    | IntConst(_) -> c
    | BoolConst(_) -> c
    | FloatConst(_) -> c
    | StringConst(_) -> c
    | Lambda(p, r, s) ->
       let new_ctx = prepare_ctx e.exp_loc (c, p) in
       let c' = bind_stmt (new_ctx, s) in
       ignore(c');
       (* We return the same contex. Local function cannot affect
        * parent enviroment. *)
       c
;;

(* Modify stmts in place, adding symbol entries to Ref terms. This musn't be
 * called more then once in a toplevel function. *)
let bind decl =
  let add_global loc t n =
    let sym = make_symbol global_ctx t n in
    try
      ignore(find_sym global_ctx.ctx_syms n);
      err (loc, xf "redefintion of global symbol `%s'" n)
    with Not_found -> Hashtbl.add globals n sym in
  match decl with
      FuncDecl(f) ->
        let fntype = Func(f.fn_ret_type,
                          (List.map (fun (t, _) -> t) f.fn_parms)) in
        add_global f.fn_loc fntype f.fn_name;
        let initial_ctx = prepare_ctx f.fn_loc (global_ctx, f.fn_parms) in
        ignore (bind_stmt (initial_ctx, f.fn_body))
    | GlobalDecl(_, t, n, loc) ->
        add_global loc t n
    | ExternC(x) ->
        let fntype = Func(x.xc_ret_type, x.xc_parms) in
        add_global x.xc_loc fntype x.xc_name
;;

(* Apply given function, of (exp -> 'a list) type, for all
 * expressions in a tree, returning flat list of results. *)
let rec for_all_exp exp_fnc s =
  match s.raw_stmt with
      Skip -> []
    | Exp(e) -> exp_fnc e
    | Return(e) -> exp_fnc e
    | ReturnVoid -> []
    | Block(l) -> (List.flatten (List.map (for_all_exp exp_fnc) l))
    | VarDef(_) -> []
;;

(* Return list of symbols, at given scoping level, that needs to put into the
 * closure becouse they are referenced in stmts. *)
let closure lev s =
  (*
  let rec unique r xxs =
    match xxs with
      x::xs ->
        if List.exists (fun y -> x = y) r then
          unique r xs
        else
          unique (x::r) xs
      | [] -> r in *)
  let need_closure = ref false in
  let rec exp_closure lambda_lev e =
    let ec = exp_closure lambda_lev in
    match e.raw_exp with
        SimpleBinOp(op, e1, e2) -> (ec e1) @ (ec e2)
      | SimpleUnOp(op, e1) -> ec e1
      | Ref(r) ->
          (match r.ref_symbol with
              None -> assert false
            | Some sym ->
                if r.ref_level > 0 then
                  (if lambda_lev > 0 then need_closure := true;
                   if sym#get_fnc_level = lev && not sym#in_closure then
                     (sym#put_into_closure; [sym])
                   else
                     [])
                else
                  [])
      | Call(e1, es) ->
          (ec e1) @ (List.flatten (List.map ec es))
      | IntConst(_) -> []
      | BoolConst(_) -> []
      | FloatConst(_) -> []
      | StringConst(_) -> []
      | Lambda(_, _, s) ->
         for_all_exp (exp_closure (lambda_lev + 1)) s in
  (!need_closure, for_all_exp (exp_closure 0) s)
;;

