(* $Id: gonttypes.ml,v 1.5 2001/11/15 21:28:53 malekith Exp $ *)

open Gontutil
open Gontbind

exception Type_error of string

(* o[x/t] *)
let rec subst (x, t) o =
  match o with
      TyVar(y) when x = y -> t
    | TyVar(_) | Int | Void | Bool | Float | String -> o
    | Func(r1, p1) -> Func(subst (x, t) r1, (List.map (subst (x, t)) p1))
;;

let rec string_of_type_list tt =
  match tt with
      [t] -> string_of_type t
    | t::ts -> (string_of_type t) ^ ", " ^ (string_of_type_list ts)
    | [] -> ""
and string_of_type t =
  match t with
      Void -> "void"
    | Int -> "int"
    | Float -> "float"
    | Bool -> "bool"
    | String -> "string"
    | Func(t, a) -> 
        "[" ^ (string_of_type_list a) ^ "] -> " ^ (string_of_type t)
    | TyVar(x) -> ("'" ^ x)
;;


let rec is_fv v t =
   match t with
      TyVar(x) -> v = x
    | Int | Void | Bool | Float | String -> false
    | Func(r, p) -> (is_fv v r) || (List.exists (is_fv v) p)
;;

(* Term unification, the standard, given-from-God, algorithm :^) 
 * sub is boolean to specify if we want to find such O, that tO = sO 
 ( (sub = false), or tO = s (sub = true). *)
let rec mgu_ext sub (o : (id * ty) list) rrs =
  match rrs with
      r::rs ->
          (match r with
              (TyVar(x), TyVar(y)) when x = y -> 
                  mgu_ext sub o rs
            | (TyVar(x), t) when not (is_fv x t) ->
                  mgu_ext sub 
                      ((x, t) :: (List.map (fun (x, tt) -> (x, subst (x, t) tt)) o))
                      (List.map (fun (t1, t2) -> (subst (x, t) t1, 
                                                  subst (x, t) t2)) rs)
            | (t, TyVar(x)) when not (sub || is_fv x t) ->
                  mgu_ext sub 
                      ((x, t) :: (List.map (fun (x, tt) -> (x, subst (x, t) tt)) o))
                      (List.map (fun (t1, t2) -> (subst (x, t) t1, 
                                                  subst (x, t) t2)) rs)
            | (Int, Int) | (Bool, Bool) | (Void, Void)
            | (Float, Float) | (String, String) -> 
                  mgu_ext sub o rs
            | (Func(r1, p1), Func(r2, p2)) ->
                  mgu_ext sub o ((r1, r2) :: 
                                 (List.map2 (fun a b -> (a, b)) p1 p2) @
                                 rs)
            | (x, y) ->
                  if sub then
                    raise (Type_error (xf "%s is not instance of %s"
                                            (string_of_type x)
                                            (string_of_type y)))
                  else
                    raise (Type_error (xf "cannot unify %s with %s" 
                                         (string_of_type x)
                                         (string_of_type y))))
                                     
    | [] -> o
;;

let mgu = mgu_ext false []
;;

(* Substitute variables in t, that are also in s, with
 * fresh variables. *)
let rec fresh_vars s t =
  match t with
      TyVar(x) -> 
        if is_fv x s then
          TyVar(seq_name ".")
        else
          t
    | Int | Void | Bool | Float | String -> t
    | Func(r, p) -> Func(fresh_vars s r, List.map (fresh_vars s) p)
;;

(* Check if t < s. *)
let subtype s t =
  try
    ignore(mgu_ext true [] [(fresh_vars s t, s)])
  with Type_error(e) -> 
    raise (Type_error (xf "%s hence %s is not instance of %s" e
                          (string_of_type t)
                          (string_of_type s)))
;;

type ty_context = {ctx_ret_type: ty}

let typecheck_call actual formal =
  let rec helper no actuall formall =
    match formall with
        (formal::formals) ->
          (match actuall with
              (actual::actuals) -> 
                (try
                   subtype actual formal
                 with Type_error s -> 
                   raise (Type_error (xf "%s in parameter %d of call" s no)));
                helper (no + 1) formals actuals
            | _ -> assert false)
      | [] -> () in
  if (List.length formal) != (List.length actual) then
    raise (Type_error (xf "there is %d formal parameters and %d actual \
                           parameters in call"
                          (List.length formal) 
                          (List.length actual)))
  else
    helper 1 actual formal
;;

let lvalue_p e =
  match e.raw_exp with
      Ref r -> ()
    | _ -> raise (Type_error "expression is not a proper lvalue")
;;

let rec typecheck_exp (g, e) =
  try
    let (g', t) =
      match e.raw_exp with
          SimpleBinOp(op, e1, e2) ->
            let (g1, t1) = typecheck_exp (g, e1) in
            let (g2, t2) = typecheck_exp (g1, e2) in
            (match op with
                IntPlus | IntMinus | IntDiv | IntMod | IntMult -> 
                  ignore(mgu [(t1, Int); (t2, Int)]);
                  (g2, Int)
              | FloatPlus | FloatMinus | FloatDiv | FloatMult -> 
                  ignore(mgu [(t1, Float); (t2, Float)]);
                  (g2, Float)
              | Assign ->
                  subtype t1 t2;
                  lvalue_p e1;
                  (g2, t1)
              )
        | SimpleUnOp(op, e) ->
            let (g', t) = typecheck_exp (g, e) in
            (match op with
                 IntNegate -> ignore(mgu [(t, Int)]); (g', Int)
               | FloatNegate -> ignore(mgu [(t, Float)]); (g', Float))
        | Ref r ->
            (match r.ref_symbol with 
                None -> assert false
              | Some sym -> (g, sym#get_type))
        | Call(f, act) ->
            let (g', t) = typecheck_exp (g, f) in
            let (g'', tact) = list_apply typecheck_exp (g', act) in
            (match t with
                Func(r, form) -> typecheck_call tact form; (g'', r)
              | _ -> raise (Type_error "called object is not a function"))
        | IntConst _ -> (g, Int)
        | BoolConst _ -> (g, Bool)
        | FloatConst _ -> (g, Float)
        | StringConst _ -> (g, String)
        | Lambda(a, r, s) -> 
            ignore(typecheck_stmt ({g with ctx_ret_type = r}, s));
            (g, Func(r, (List.map (fun (t, n) -> t) a))) in
    e.exp_ty <- Some t;
    (g', t)
  with Type_error s -> raise (Syntax_error (e.exp_loc, s))

and typecheck_stmt (g, s) =
  try
    match s.raw_stmt with
        Skip -> 
          (g, Void)
      | Exp(e) -> 
          let (g', _) = typecheck_exp (g, e) in 
          (g', Void)
      | Return(e) -> 
        let (g', t) = typecheck_exp (g, e) in
        subtype g.ctx_ret_type t; 
        (g', Void)
      | ReturnVoid ->
        subtype g.ctx_ret_type Void;
        (g, Void)
      | Block(ss) ->
        let (g', _) = list_apply typecheck_stmt (g, ss) in
        (g', Void)
      | VarDef(_) ->
        (g, Void)
  with Type_error e -> raise (Syntax_error (s.stmt_loc, e))
;;

let typecheck_decl d =
  match d with
      FuncDecl f -> 
        ignore(typecheck_stmt ({ctx_ret_type = f.fn_ret_type}, f.fn_body))
    | GlobalDecl (_, _, _, _) -> ()
    | ExternC _ -> ()
;;

let type_of e =
  match e.exp_ty with
      Some t -> t
    | None -> assert false
;;
