(* $Id: gontutil.mli,v 1.4 2001/11/15 19:01:08 malekith Exp $ *)

(* Identifier. *)
type id = string

type location = {mutable loc_file: string; mutable loc_line: int}

(* Types used by gont compiler. *)
type ty =
    Int                         (* integer of the same size as pointer *)
  | Void                        (* unit type *)
  | Bool                        (* boolean, same size as Int *)
  | Float                       (* C's double *)
  | Func of ty * (ty list)
  | String
  | TyVar of id
;;

class type symbol = 
  object
    method get_name : id
    method get_type : ty
    method get_fnc_level : int
    method get_scope_level : int
    method in_closure : bool
    method put_into_closure : unit
  end
  
type op2 =
    IntPlus
  | IntMinus
  | IntMult
  | IntDiv
  | IntMod
  | FloatPlus
  | FloatMinus
  | FloatMult
  | FloatDiv
  | Assign
  
type op1 =
    IntNegate
  | FloatNegate

type raw_exp =
    SimpleBinOp of op2 * exp * exp
  | SimpleUnOp of op1 * exp
  | Ref of ref_record   (* reference to variable *)
  | Call of exp * exp list
  | IntConst of string
  | BoolConst of bool
  | FloatConst of string
  | StringConst of string
  | Lambda of (ty * id) list * ty * stmt
and exp = {mutable raw_exp: raw_exp; 
           mutable exp_ty: ty option; 
           exp_loc: location}
and ref_record = {ref_name: id;
                  mutable ref_level: int;
                  mutable ref_symbol: symbol option}
and raw_stmt =
    Skip
  | Exp of exp
  | Return of exp
  | ReturnVoid
  | VarDef of var_record
  | Block of stmt list
and stmt = {raw_stmt: raw_stmt;
            stmt_loc: location}
and var_record = {var_name: id;
                  var_type: ty;
                  mutable var_symbol: symbol option}

type storage =
    Local
  | Public

type fn_decl = {fn_storage: storage;
                fn_ret_type: ty; 
                fn_name: id; 
                fn_parms: (ty * id) list;
                fn_body: stmt;
                fn_loc: location}

type extern_c_decl = {xc_ret_type: ty; 
                      xc_name: id; 
                      xc_parms: ty list;
                      xc_loc: location}
type decl =
    FuncDecl of fn_decl
  | GlobalDecl of storage * ty * id * location
  | ExternC of extern_c_decl

;;

exception Syntax_error of location * string
exception Eof
;;

(* Utils. *)

(* Return unique name prefixed with string. String should start
 * with '@', unless it is going to get exported. 
 *)
val seq_name : id -> id

(* Output debugging (trace) message. *)
val debug : string -> unit

(* An alias for Printf.sprintf. *)
val xf : ('a, unit, string) format -> 'a

(* Display warning. *)
val warn : location * string -> unit

(* Display error message. *)
val err_msg : location * string -> unit

(* Raise Syntax_error. *)
val err : location * string -> 'a

(* 'a is type for context, 'b is input value, and 'c is output value. *)
val list_apply : ('a * 'b -> 'a * 'c) -> 'a * 'b list -> 'a * 'c list
;;
